import pygame, sys, os, random, math, Queue, time
from operator import itemgetter, attrgetter

#Width of a standard game block
BLOCK_WIDTH = 32
X_OFFSET = 20
Y_OFFSET = 48

def pixelToGrid(p):
    x = (p[0] - X_OFFSET)/float(BLOCK_WIDTH)
    y = (p[1] - Y_OFFSET)/float(BLOCK_WIDTH)
    return [x,y]

def pixelToIntGrid(p):
    pos = pixelToGrid(p)
    return [ int(pos[0]),int(pos[1]) ]

def gridToPixel(p):
    return [p[0]*BLOCK_WIDTH+X_OFFSET,p[1]*BLOCK_WIDTH+Y_OFFSET]

#Types of DMG
#DMG is for a direct damage type
#SAT is for a Satilite type, found in the tower class
DMG_PHYSICAL = 0
DMG_ACID     = 1
DMG_HACKING  = 2
DMG_EMP      = 3
DMG_SPEED    = 4
DMG_MAGNETIC = 5
#SAT_MICROWAVE= 6


#A base class for game objects
class Entity(pygame.sprite.Sprite):

    
    #Color = RGB color list indicating the color of the object
    #ProjC0lor = Stores what the Enitity's attack looks like
    #maxHP = an integer representing the total amount of hp the object possesses
    #currentHP = an integer representing the current hp of the unit
    #suscept = a list of integers indicating resistances 
    #attackPower = an integer representing the attack power of the unit
    #attackType = the type of attack, represented by a string
    #attackRange = an integer representing the maximum number of blocks that the enitity can attack in
    #AoE = Can this unit do damage over an area
    #passable = a boolean, can units walk over this unit
    #frozen = boolean of frozen status
    #dying = boolean of whether or not the object is dying
    #team = a string with the team name of the entity
    #cooldown = an integer representing the required time taken between shots of damage
    #currentCooldown = the current cooldown time of the entity before it can take an attack action
    #width = an integer representing the amount of blocks that the object occupies
    def __init__(self, icon, projColor, hp, susceptibility, attackPower, attackType,\
                 attackRange, attackSpeed, AoE, passable, team, cooldown, topLeftPos,\
                 width, descr,cost):
        pygame.sprite.Sprite.__init__(self)
        self.projColor = projColor
        self.maxHP = hp
        self.currentHP = hp
        self.suscept = susceptibility
        self.attackPower = attackPower
        self.attackType = attackType
        self.attackRange = attackRange
        self.attackSpeed = attackSpeed
        self.AoE = AoE
        self.passable = passable
        self.frozen = 0
        self.dying = False
        self.team = team
        self.cooldown = cooldown
        self.currentCooldown = 0
        self.width = width
        self.descr = descr
        self.enabled = Entity.IMG_ATTACKING
        self.cost = cost
        
        #Sprite Parameters
        self.icon=pygame.image.load(icon)
        self.iconName = icon
        self.createImages()
        self.createRangeImage()

        self.image=self.Aimage
        self.rect = self.image.get_rect()
        self.rect.topleft = [topLeftPos[0]*BLOCK_WIDTH+X_OFFSET,topLeftPos[1]*BLOCK_WIDTH+Y_OFFSET]

    def createRangeImage(self):
         #disabled - mask1
         w = int(2*BLOCK_WIDTH * self.attackRange)
         p = [w/2,w/2] 
 
         self.rangeImg = pygame.Surface([w,w],pygame.SRCALPHA, 32)
         self.rangeImg.fill((0,0,0,0))
         pygame.draw.circle(self.rangeImg,(200,200,200,175),p,int(w/2))
         
 
    def drawRange(self,surf):
         Gpos =self.getPos()
         Gpos = (Gpos[0]-self.attackRange,Gpos[1]-self.attackRange)
         surf.blit(self.rangeImg,gridToPixel(Gpos) )

    def createImages(self):
        
        #normal icon
        self.Aimage = pygame.Surface([BLOCK_WIDTH * self.width, BLOCK_WIDTH * self.width],pygame.SRCALPHA, 32)
        self.Aimage.fill((0,0,0,0))

        #draw background                 
        p = [int(BLOCK_WIDTH/2 * self.width),int( BLOCK_WIDTH/2 * self.width)]
        w = int(BLOCK_WIDTH/2 * self.width) 
        pygame.draw.circle(self.Aimage,self.team,p,w)

        #draw icon
        self.Aimage.blit(self.icon,(0,0))

        #disabled - mask1
        img = pygame.Surface([BLOCK_WIDTH * self.width, BLOCK_WIDTH * self.width],pygame.SRCALPHA, 32)
        img.fill((0,0,0,0))
        pygame.draw.circle(img,(0,0,0,175),p,w)
        #draw disabled icon
        self.Dimage = pygame.Surface([BLOCK_WIDTH * self.width, BLOCK_WIDTH * self.width],pygame.SRCALPHA, 32)
        self.Dimage.blit(self.Aimage,(0,0))
        self.Dimage.blit(img,(0,0))


        #waiting - mask2
        img = pygame.Surface([BLOCK_WIDTH * self.width, BLOCK_WIDTH * self.width],pygame.SRCALPHA, 32)
        img.fill((0,0,0,0))
        pygame.draw.circle(img,(0,0,0,75),p,w)

        #draw waiting icon
        self.Wimage = pygame.Surface([BLOCK_WIDTH * self.width, BLOCK_WIDTH * self.width],pygame.SRCALPHA, 32)
        self.Wimage.blit(self.Aimage,(0,0))
        self.Wimage.blit(img,(0,0))
        self.setImageState(Entity.IMG_ATTACKING)        

    def addPolicy(self,p):
        self.policy = p
        p.registerRobot(self)


    def getCost(self):
        return self.cost


    IMG_ATTACKING = 0
    IMG_WAITING = 1
    IMG_DISABLED = 2


    def setImageState(self,val):
        self.enabled = val
        if(self.enabled == Entity.IMG_ATTACKING):
            self.image=self.Aimage
        elif (self.enabled == Entity.IMG_WAITING):
            self.image=self.Wimage
        elif(self.enabled == Entity.IMG_DISABLED):
            self.image=self.Dimage
    
    #changes the state of the entity based upon the attackers attributes
    def takeDamage(self,dmg,dmgType,source):
        self.setImageState(Entity.IMG_ATTACKING)
        source.setImageState(Entity.IMG_ATTACKING)
        if dmgType == DMG_EMP:
           theoretic =  int(dmg * self.suscept[dmgType])
           #frozen does not stack
           if theoretic >= self.frozen:
               self.frozen = theoretic
        elif dmgType == DMG_HACKING:
            prob = (dmg * self.suscept[dmgType]) * 0.01
            if prob > random.uniform(0,2):
                self.team = source.team
                if(isinstance(self,Tower)):
                    self.policy.generatePolicy()
                    source.polocy.generatePolicy()            
                self.createImages()
                self.setImageState(Entity.IMG_ATTACKING)
                self.addPolicy(source.policy)
        else:
            self.currentHP = self.currentHP - (dmg * self.suscept[dmgType])
            if self.currentHP > self.maxHP:
                self.currentHP = self.maxHP
                
    #Tells an enemy to take damage using the takeDamage method
    def giveDamage(self,enemy):
        enemy.takeDamage(self.attackPower,self.attackType,self)
        #MKB cooldown starts at 0
        self.currentCooldown =  0

    #Checks if the entity is alive
    def isAlive(self):
        return not self.dying
        
    #Accesses the team of the entity
    def getTeam(self):
        return self.team

    #Returns a tuple representing the center of the entity
    def getLocation(self):
        return self.getPos()

    def getPos(self):
        return pixelToGrid(self.rect.center)

    #Finds the distance from the entity to another unit 
    def findDistance(self,unit):
        myLocation = self.getLocation()
        unitLocation = unit.getLocation()
        dx = (unitLocation[0] - myLocation[0])**2
        dy = (unitLocation[1] - myLocation[1])**2
        return  (dx + dy)**.5

    def getRange(self):
        return self.attackRange

    def moveBy(self,p):
        self.rect.move_ip(p[0]*BLOCK_WIDTH,p[1]*BLOCK_WIDTH)

    def moveTLTo(self,p):
        self.rect.topleft = gridToPixel(p)
        #print "wanted",p,"got",self.getPos()

    def moveBySmall(self,start,iteration,vec):
        x = start[0] + iteration*vec[0]
        y = start[1] + iteration*vec[1]
        self.rect.center = gridToPixel( (x,y) )

    def moveToPixel(self,p):
        self.rect.topleft = p

    def recenter(self):
        p = self.getPos()
        dx = (math.ceil(p[0]) + math.floor(p[0]))/2.0 - p[0]
        dy = (math.ceil(p[1]) + math.floor(p[1]))/2.0 - p[1]
        self.moveBy((dx,dy))
        #print self.getPos()

    def update(self,*args):
        #self.showHealth(args[2])
        """Returns true if the Entity should take an action"""
        if(self.attackPower ==0 or \
           (self.enabled==Entity.IMG_ATTACKING and self.currentCooldown<.25*self.cooldown) ):
            self.setImageState(Entity.IMG_ATTACKING) #overwrite in subclass
        else:
            self.setImageState(Entity.IMG_WAITING) #overwrite in subclass

        #print "getting called",
        if self.frozen > 0:
            self.frozen-=1
        else:
            self.currentCooldown += 1
            if(self.currentCooldown >= self.cooldown):
                self.currentCooldown = 0
                return True
        return False

     #returns a boolean, can this unit be passed over
    def  isPassable(self):
        return self.passable
            
    def isAlive(self):
        return not self.dying
    
    def getTeam(self):
        return self.team

    def drawDescription(self, surf, x, y ):
        font = pygame.font.Font("ARNPRIOR.TTF",16)
        text = self.descr + "\n" +"HP - " + str(self.maxHP) + "\n" + "Power - "\
               + str(self.attackPower) + "\n" + "Range - " + str(self.attackRange)\
               + "\n" + "Speed - " + str(self.attackSpeed) + "\n" + "\n" \
               "COST - " + str(self.cost)
        messages = text.split("\n")
        for message in messages:
            messageSurface = font.render(message,False,(0,0,0))
            surf.blit(messageSurface,(x,y))
            y = y + 20

    def containsPoint(self,p):
        """ p must be in pixel coords"""
        return self.rect.collidepoint(p)

class Projectile(pygame.sprite.Sprite):

    def __init__(self,color, offset, target, source, speed, width):
        pygame.sprite.Sprite.__init__(self)
        self.target = target
        self.source = source
        self.speed = speed
        self.width = width

        #Sprite Parameters 
        self.image = pygame.Surface([BLOCK_WIDTH * self.width, BLOCK_WIDTH * self.width])
        self.image.fill(color)
        self.rect = self.image.get_rect()
        self.rect.center = gridToPixel(self.source.getPos())
 
    def getPos(self):
         x = (self.rect.center[0] - X_OFFSET)/float(BLOCK_WIDTH)

         y = (self.rect.center[1] - Y_OFFSET)/float(BLOCK_WIDTH)

         return [x,y]

    def targetPos(self):
        return self.target.getPos()

    def moveBy(self,p):
        self.rect.move_ip(p[0]*BLOCK_WIDTH,p[1]*BLOCK_WIDTH)

    
    #Finds the distance from the entity to another unit 
    def findDistance(self,unit):
        myLocation = self.getPos()
        unitLocation = unit.getPos()
        distance = math.sqrt(((unitLocation[0] - myLocation[0])**2) + math.sqrt((unitLocation[1] - myLocation[1])**2))
        return distance

    def update(self, *args):
        targetPos = self.targetPos()
        distance = self.findDistance(self.target)
        Dx = (targetPos[0] - self.getPos()[0])
        Dy = (targetPos[1] - self.getPos()[1])
        if distance < self.speed:
            self.moveBy((Dx,Dy))
            self.source.giveDamage(self.target)
            self.kill()
        else:
            displacement = [self.speed * Dx/distance,self.speed * Dy/distance]
            self.moveBy(displacement)

class Healthbar(pygame.sprite.Sprite):

    def __init__(self, source):
        pygame.sprite.Sprite.__init__(self)
        self.source = source
        self.image = pygame.Surface([20,5])
        self.image.fill((0,250,0))
        self.rect = self.image.get_rect()
        self.image.get_rect()
        pos = gridToPixel(self.source.getPos())
        self.rect.topleft = [pos[0]-25,pos[1]-35]
        
    def update(self, *args):
        coords = self.source.getPos()
        if self.source.currentHP <= 0:
            self.kill()
            return
        width = (float(self.source.currentHP)/float(self.source.maxHP)) * 20
        self.image = pygame.Surface([width,5])
        self.image.fill((250,0,0))
        self.rect = self.image.get_rect()
        self.image.get_rect()
        pos = gridToPixel(self.source.getPos())
        self.rect.topleft = [pos[0]-12,pos[1]-18]

        #Tower class derived from the entity class, represents a tower on the game grid
class Tower(Entity):

        #isSatilite = a boolean, is this tower used for deploying Satilites
        #Satilite = a storage compartment for a Satilite object, its empty by default
        def __init__(self, color, projColor, hp, susceptibility, attackPower,\
                     attackType, attackRange, attackSpeed, AoE, passable, team, \
                     cooldown, topLeftPos, width, cost, descr):
            Entity.__init__(self, color, projColor, hp, susceptibility, \
                            attackPower, attackType, attackRange, attackSpeed, \
                            AoE, passable, team, cooldown, topLeftPos, width, descr,cost)
            #self.cost = cost
            #self.isSatilite = isSatilite
            #if self.isSatilite:
             #   self.currentCooldown = self.cooldown
            #self.Satilite = None

        def clone(self):
            return Tower(self.iconName,self.projColor,self.maxHP, self.suscept, \
                            self.attackPower, self.attackType, self.attackRange, self.attackSpeed, \
                            self.AoE, self.passable, self.team, self.cooldown, \
                            pixelToGrid(self.rect.topleft),self.width, self.cost,self.descr)


        def update(self,*args):
            if self.currentHP <= 0:
                self.image.fill( (200,200,0) )
                self.kill()
            elif super(Tower,self).update(args) :
                enemies = []
                for i in args[1]:
                    if(i.team != self.team):
                        enemies.append(i)

                projects = self.attack(enemies)
                #print projects
                if projects != []:
                    args[0].add(projects)
                else:  #"store" cooldown so tower will instantlt fire when robot comes into range
                    self.currentCooldown=self.cooldown
            else:
               return


        #If the tower can take a turn, it will attack
##        def takeTurn(self, units):
##            if self.isAlive():
##                if self.update():
##                    self.attack(units)
##                else:
##                    return

        #self,image, offset, target, source, speed      
        #The Tower will either attempt an AoE attack or a bolt attack depending upon its type
        #If a unit is not in range, it will not attack, and will not put up a cooldown
        def attack(self, units):
            projectiles = []
            if len(units) == 0:
                return []
            #Attacks using linear attack methods
            if self.AoE == False and self.passable == False:
                target = self.shotAttack(units)
                if target:
                    return Projectile(self.projColor,(0,0),target,self,self.attackSpeed,.2)
##            If the tower has a  Satilite, uses the Satilites power
##            elif self.Satilite != NULL:
##                if self.Satilite.getCharges() <= 0:
##                    self.dying = True
##                    return
##                self.useSatilite(units)
            #Uses an AoE attack 
            elif self.AoE:
                targeted = self.areaAttack(units)
                if targeted:
                    for target in targeted:
                        newProjectile = Projectile(self.projColor,(0,0),target,self,self.attackSpeed,.4)
                        projectiles.append(newProjectile)
                        #print projectiles
                        return projectiles
            #attacks enemys standing on the tower
            elif self.passable:
                #print "Trap"
                targeted = self.areaAttack(units)
                if targeted:
                    for target in targeted:
                        self.giveDamage(target)
                    self.currentHP = 0
                    self.kill()
            #print projectiles
            return projectiles
                        
        #The tower will find the closest unit and return it
        def shotAttack(self, units):
            currentTarget = units[0]
            for unit in units:
                if self.findDistance(unit) < self.findDistance(currentTarget):
                    currentTarget = unit
            if self.findDistance(currentTarget) > self.attackRange:
                return []
            return currentTarget

        #The tower will return a list of all units within range
        def areaAttack(self, units):
            #print "Called areaAttack"
            currentTargets = []
            for unit in units:
                #print str(self.findDistance(unit)) + " myrange - " + str(self.attackRange)
                if self.findDistance(unit) < self.attackRange:
                    currentTargets.append(unit)
            return currentTargets

        #Gives the tower a Satilite Object
##        def deploySatilite(self):
##            self.Satilite = Satilite(self.attackType)

        #Uses the Satilites ability
##        def useSatilite(self, units):
##            self.Satilite.fire()

        #A subclass stored in a tower class capable of creating a satilite
##        class Satilite(object):
##
##            #attackType = Stores the attack type of the Satilite, this represents the type of Satilite 
##            #charges =  the amount of the times its attack or effect can be used before it destroys the tower it spawns from
##            def __init__(self, attackType):
##                self.attackType = attackType
##                self.charges = 0
##                if self.attackType == 6:
##                    self.charges = 1
##
##            #getter for the Satilite's charge
##            def getCharges(self):
##                return self.charges
##
##            #uses the Satilite's ability and decrements its charge
##            def fire(self):
##                if self.attackType == 6:
##                    self.charges = self.charges - 1
##                    return

class Robot(Entity):

    def __init__(self, color, projColor, hp, susceptibility, attackPower, attackType, attackRange, attackSpeed, AoE, team, cooldown, topLeftPos, width):
        Entity.__init__(self, color, projColor, hp, susceptibility, \
                        attackPower, attackType, attackRange, attackSpeed,\
                        AoE, True, team, cooldown, topLeftPos, width, "I'm a Robot",5)
        self.vector = (0,0)
        self.deathTimer = 5
        self.state = "waiting"
        self.game = None
        self.iter=0
        self.lastPos = self.getPos()

    def setGame(self,game):
        self.game=game

    def multHP(self,mod):
        self.maxHP *=mod
        self.currentHP *=mod
        
        
   


    def attackEntity(self,entity):
        #am I close enough, then move closer
        #thing this should work with enemy bots as well
        d = self.findDistance(entity)
        if(d > (self.attackRange +entity.width) ):
            pos = self.getPos()
            center = entity.getPos()
            dx = center[0] - pos[0]
            dy = center[1] - pos[1]
            distToCover = d - entity.width-self.attackRange

            self.state="walking"            
            self.vector = ( distToCover*dx/(d*self.cooldown) ,distToCover*dy/(d*self.cooldown) )
            self.moveBy(self.vector)
        else:
            self.state="attacking"
            self.giveDamage(entity)
            #print "ATTACKING ",entity.descr




    def attackEnemyBots(self,bots):

        enemyBots = []
        for b in bots:
            if(b.team != self.team):
                enemyBots.append(b)

        for b in enemyBots:
            if(self.findDistance(b)<2*self.attackRange):
                self.attackEntity(b)
                return True
        return False
                



    
    def update(self,*args):
        bots = args[0]
        
        if(self.state=="dying"):
            self.deathTimer = self.deathTimer - 1
            if(self.deathTimer <=0):
                self.game.addMoney(self.cost,self.maxHP)
                self.kill()
        
        #print self.currentHP
        elif self.currentHP <= 0:
            self.setImageState(Entity.IMG_DISABLED)
                               
            self.state="dying"

        #print "updating",self.vector
        elif( super(Robot,self).update(args) ):

            #if there is an enemy bot in within range 1 I should attack it
            if(self.attackEnemyBots(bots)):
                return True
            
            #plan out the next action
            v = self.policy.getDirection(self.getPos())
            if(v==None):
                return True
            nextPos = self.getPos()[0]+v[0],self.getPos()[1]+v[1]
            t = self.policy.getTower(nextPos)
            if(t != None and not t.isPassable()):
                self.attackEntity(t)
                return True
            
            #else we will continue walking            
            self.state="walking"
            self.lastPos = self.getPos()
            self.iter=1
            self.vector = (v[0]/float(self.cooldown),v[1]/float(self.cooldown))
            self.moveBySmall(self.lastPos,self.iter,self.vector)
            return True
        elif(self.state=="walking"):
            #move by vector
            self.iter+=1
            #print "walked from",self.rect.center,
            self.moveBySmall(self.lastPos,self.iter,self.vector)
        return False

REWARD_DISCOUNT = .99999
BASE_REWARD = 1000
TOWER_REWARD = 0

class PolicyCell(object):

    
    def __init__(self,pos,cellsize):
        self.setCellSize(cellsize)
        self.pos = pos
        self.dir = (0,1)
        self.reward =0
        self.eReward=0
        self.timeToEnter = 1
        self.damagePerRound = 0
        self.tower = None
        

    def addDamage(self,tower,target,remove=False):
        #print "-------------"
        r=1
        if(remove):
            r = -1
        x = self.pos[0]
        y = self.pos[1]
        samples = [ (x+self.quarter,y+self.quarter), \
                    (x+self.trier,y+self.quarter), \
                    (x+self.quarter,y+self.trier), \
                    (x+self.trier,y+self.trier), \
                    (x+self.half,y+self.half) ]
        s = 0
        cx,cy = tower.getPos()
        dist = tower.getRange()**2
        for x,y in samples:
            d = ((x-cx)**2 + (y-cy)**2)
            #print x,y,cx,cy,d,dist,
            if(  d <=  dist):
                #print "yes"
                s+=1
            #else:
                #print "no"
        self.damagePerRound += r*(s/5.0)* (tower.attackPower * target.suscept[tower.attackType] * target.cooldown/float(tower.cooldown) )
        

    def calcTimeToEnter(self,tower,target,remove=False):
        if(remove):
            self.timeToEnter=1
            self.tower=None
        else:
            #print tower.descr
            attackPower = target.attackPower
            if(target.attackPower ==0):
                attackPower = 1000
            if(not tower.isPassable() and tower.attackPower>0  ):
                self.timeToEnter =  tower.currentHP/(attackPower * tower.suscept[target.attackType])
            else:
                self.timeToEnter = 1
            self.tower = tower
            

    
    def getTower(self):
        return self.tower

    def setTower(self,t):
        self.tower=t
    
    def setCellSize(self,size):
        self.cellSize = size
        self.quarter = size*.15
        self.half = size*.5
        self.trier = size*.85

    def __str__(self):
        #return "["+str(self.dir)+ "%5d %4d %2d"%(self.eReward,self.damagePerRound,self.timeToEnter)+"]"
        return "["+str(self.pos)+" "+str(self.timeToEnter)+" "+str(self.dir)+ "R%4d D%3d"%(self.eReward,self.damagePerRound)+"] "


    def calcRewardTo(self,n):
        return REWARD_DISCOUNT*n.eReward - self.damagePerRound*n.timeToEnter

    def choosePath(self, N):
        if(len(N)==0):
            self.dir=(0,0)
            return False
        #print self.pos
        
        maxN = N[0]
        maxReward = self.calcRewardTo(maxN)
        for i in N:
            c = self.calcRewardTo(i)
            #print "   ",i.pos,c,i.timeToEnter
            if(c > maxReward):
                maxN = i
                maxReward = c

        #print self.pos,maxN.pos
        d = (maxN.pos[0]-self.pos[0],maxN.pos[1]-self.pos[1])
        #print "   ",d,self.dir
        ret = d !=self.dir
        self.dir =d
        self.eReward = maxReward + self.reward #- self.damagePerRound
        return ret

    def addReward(self,rew,remove=False):
        r=1
        if(remove):
            r = -1
        self.reward += r*rew
        self.eReward +=r*rew

class Policy(pygame.sprite.Sprite):

    up = pygame.image.load('up.png')
    down = pygame.image.load('down.png')
    left = pygame.image.load('left.png')
    right = pygame.image.load('right.png')

    def __init__(self,cols,rows,cellsize,target,towers,dest):
        pygame.sprite.Sprite.__init__(self)
        self.cols  = int(cols*(1.0/cellsize))
        self.rows = int(rows*(1.0/cellsize))
        self.target=target
        self.map = []
        self.cellsize = cellsize
        self.dest=dest
        for i in range(self.cols):
            col = []
            for j in range(self.rows):
                col.append(PolicyCell((i*cellsize,j*cellsize),cellsize))
            self.map.append(col)

        self.calcDamageAndRewards(towers,dest)
        self.generatePolicy()
        self.Users = pygame.sprite.Group()
        #print self

    def drawDirs(self,surf):

        for i in range(self.cols):

            for j in range(self.rows):

                d = self.map[i][j].dir

                if(d==(0,-1)):

                    surf.blit(Policy.up,gridToPixel( (i,j) ) )

                elif(d==(1,0)):

                    surf.blit(Policy.right,gridToPixel( (i,j) ) )

                elif(d==(-1,0)):

                    surf.blit(Policy.left,gridToPixel( (i,j) ) )

                elif(d==(0,1)):

                    surf.blit(Policy.down,gridToPixel( (i,j) ) )

     
    def myupdate(self,*args):
        #print "Hi there ",args,len(args)
        if(len(self.Users.sprites()) == 0):
            #print "killing Policy"
            self.kill()
            return
        #self.clearExpected()
        add,remove = args
        for a in add:
            self.addTower(a)
        for a in remove:
            self.addTower(a,True)
        self.generatePolicy()

    def clearExpected (self):
        for col in self.map:
            for cell in col:
                cell.eReward=cell.reward

    def registerRobot(self,r):
        self.Users.add(r)

    def __str__(self):
        ret = ""
        for row in range(8,self.rows):#self.rows):
            s =""
            for col in range(8,self.cols):
                s+="%20s"%(str(self.map[col][row]))
            ret+=s+"\n"
        return ret

    def getPos(self,x,y):
        """returns a list of cells that are part of this grid cell"""
        cells = []
        c = int(1.0/self.cellsize)
        for i in range(int (x*c),int(x*c+c)):
            for j in range(int (y*c),int(y*c+c)):
                #print str(i) + " " + str(j)
                cells.append(self.map[i][j])
        return cells

    def getDirection(self,pos):
        x = int (pos[0]/self.cellsize)
        y = int (pos[1]/self.cellsize)
        #print x,y,pos
        #print "     ",self.map[x][y].dir

        return self.map[x][y].dir

    def getCardNeighbors(self,pos):
        N = []
        x = pos[0]
        y = pos[1]
        if(x>0):
            N.append(self.map[x-1][y])
        if(x<self.cols-1):
            N.append(self.map[x+1][y])
        if(y>0):
            N.append(self.map[x][y-1])
        if(y<self.rows-1):
            N.append(self.map[x][y+1])
        return N

    def filterFriendlies(self,N):
        a =[]
        for i in N:
            if(i.tower == None or i.tower.team != self.target.team \
               or i.tower.isPassable()):
                a.append(i)
        return a


    def calcDamageAndRewards(self,towers,dest):
        for cell in self.getPos(dest[0],dest[1]):
            cell.addReward(BASE_REWARD)

        for t in towers:
            x,y = t.getPos()
            for cell in self.getPos(x,y):
                cell.addReward(TOWER_REWARD)
                cell.calcTimeToEnter(t,self.target)
            self.calcDamage(t)


    def calcDamage(self,tower,remove=False):
        """this can be speed up"""
        if(tower.team == self.target.team):
            return
        for x in range(self.cols):
            for y in range(self.rows):
                self.map[x][y].addDamage(tower,self.target,remove)

    def addTower(self,tower,remove=False):
        x,y = tower.getPos()
        for cell in self.getPos(x,y):
            cell.addReward(TOWER_REWARD,remove)
            cell.calcTimeToEnter(tower,self.target,remove)
        self.calcDamage(tower,remove)

    def generatePolicy(self):
        #print self
        self.clearExpected()
        for i in range(100):
#            print "pass",i
            changes=False
            for i in range(self.cols):
                for j in range(self.rows):
                    if i==self.dest[0] and j == self.dest[1]:
                        self.map[i][j].dir=(0,0)
                        continue
                    if(self.map[i][j].choosePath(self.filterFriendlies(self.getCardNeighbors( (i,j) )))):
                        changes=True
            if(not changes):
                #print "Policy found in %d rounds"%(i)
                #print self
                return
            
    def getTower(self,pos):
        t = self.map[int(pos[0])][int(pos[1])].getTower()

        if(t != None):
            if(t.currentHP <= 0):
               self.addTower(t,True)
               self.generatePolicy()
               return None
        return t

class towerFactory(object):

    def __init__(self):
        self.towerData = self.loadLibrary("towers2.csv")

    def loadLibrary(self,filename):
        t = open(filename)
        lines = t.readlines()
        t.close()
        data = {}
        for line in lines[1:]:
            short = line.strip()
            if( len(short)>0 and short[0] != '#'):
                            
                words = line.split(';')
                data[words[0]] = words[1:]
                #print words[0],words[1:]
        return data

    def makeArgs(self,data):
        args=""
        for i in data:
            args+=i+" , "
        return args[:-2]

    def getTower(self,tower,pos):
        #print self.getKeys()
        
        data = self.towerData[tower]
        cmd = "t = Tower("+self.makeArgs(data)+")"
        exec(cmd)
        t.moveTLTo(pos)
        #print "worked off", tower
        return t

    def getKeys(self):
        return self.towerData.keys()




class robotFactory(object):

    def __init__(self):
        self.robotData = self.loadLibrary("robots2.csv")

    def loadLibrary(self,filename):
        t = open(filename)
        lines = t.readlines()
        t.close()
        data = {}
        for line in lines[1:]:
            short = line.strip()
            if( len(short)>0 and short[0] != '#'):
                            
                words = line.split(';')
                data[words[0]] = words[1:]
                #print words[0],words[1:]
        return data

    def makeArgs(self,data):
        args=""
        for i in data:
            args+=i+" , "
        return args[:-2]

    def getRobot(self,t,pos,hpmod=1):
        data = self.robotData[t]
        cmd = "t = Robot("+self.makeArgs(data)+")"
        exec(cmd)
        t.moveTLTo(pos)
        t.multHP(hpmod)
        return t

class Wave(object):

    def __init__(self,rtype,count,rate,nextWave):
        self.rType = rtype
        self.count = count
        self.rate = rate
        self.rate_counter=0
        self.nextWave = nextWave


    def doAction(self):
        "returns the name of a robot if one should be sent out, None otherwise"
        if(self.count <= 0):
            self.nextWave-=1
            return None
        elif(self.rate_counter >=self.rate):
            self.rate_counter=0
            self.count -=1
            return self.rType
        else:
            self.rate_counter+=1
            return None
         

    def isDone(self):
        return self.nextWave <=0


        

class Game(object):

    HOVER = "hover"
    SELECTED="selected"
    POSITIONING ="positiononing"
    
    def __init__(self):
        self.policies = pygame.sprite.Group()
        self.surf = pygame.display.set_mode((640,480))
        self.bg = pygame.image.load("bgpic.jpg")
        self.rF = robotFactory()
        self.tF = towerFactory()
        self.width  = 12
        self.height = 10
        self.towers = pygame.sprite.RenderUpdates()
        self.robots = pygame.sprite.RenderUpdates()
        self.projectiles = pygame.sprite.RenderUpdates()
        self.healthbars = pygame.sprite.RenderUpdates()
        self.purchase = []

        self.state=Game.HOVER
        self.focus = None
        self.newTower = None
        self.money = 350

        self.waves =[]
        
        self.waveNumber=0
        self.score =0
        self.showPolicy =True
        self.showBullets =True
        self.showGrid =True
        self.showHealth =True
        self.towerPolicy=None
        self.robotPolicy = None
        self.lastPolicy = None
        
        self.base = self.addTower("base",(0,0))
        self.launch=self.addTower("launch",(self.width-1,self.height-1))
        
        for i in range(5):
            self.addTower("laser",(i,1))
##        #self.addTower('hack',(self.width-2,self.height-2))

        self.towerPolicy = Policy(self.width,self.height,1,self.base,self.towers.sprites(),self.launch.getPos())
        self.policies.add(self.towerPolicy)
        for i in self.towers.sprites():
            i.addPolicy(self.towerPolicy)

        self.generateTowerButtons(440,90,4)
            
        self.generateEnemys()
        self.gColor=pygame.Color(0,0,0)

    def addMoney(self,money, hp):
        self.money +=money
        self.score +=int(hp/10)
        
    def addTower(self,text,pos):
        t = self.tF.getTower(text,pos)
        return self.plantTower(t)

        
    def plantTower(self,tower):
        newBar = Healthbar(tower)
        self.healthbars.add(newBar)
        self.towers.add(tower)
        if(self.towerPolicy != None):
            tower.addPolicy(self.towerPolicy)
        for p in self.policies.sprites():
            p.myupdate( [tower],[] )
            #p.myupdate(self.towers.sprites())
        return tower


    def addRobot(self,text,pos):
        t = self.rF.getRobot(text,pos)
        t.setGame(self)
        self.robots.add(t)
        for p in self.policies.sprites():
            p.myupdate( [t],[] )
        
    def generateEnemys(self):
        options = ["walker","runner","brute","ranger"]
        options2 =["walker","runner","ranger"]
        for i in range(20):
            if(i<4):
                t = options2[random.randint(0,len(options2)-1)]
            else:
                t = options[random.randint(0,len(options)-1)]
            if(t == 'brute'):
                w = Wave(t,3,12,90)
            else:
                w = Wave(t,6,12,90)
            self.waves.append(w)
##
        w = Wave("brute",6,30,200)
        self.waves.append(w)
            
##            x = random.randint(0,self.width-1)
##            y = random.randint(0,self.height-1)
##            pos = (x,y)
##
##            #self.robots.add( self.addRandomBot( (255,255,0),"red") )
##            self.addRobot(,pos)
##        P = Policy(self.width,self.height,1,self.robots.sprites()[0],self.towers.sprites(),(0,0))
##        self.policies.add(P)
##        for i in self.robots.sprites():
##            i.addPolicy(P)
##        #self.surf.fill((0,0,0)


    def generateTowerButtons(self,xoff,yoff,rowmax):
        keys = self.tF.getKeys()
        for key in keys:
            t = self.tF.getTower(key,(0,0))
            if(t.getCost()>0):
                self.purchase.append(t)
                t.addPolicy(self.towerPolicy)
        self.purchase.sort(key=attrgetter('cost'))
        #now set positions
        x = 0
        y = 0
        margin = 10

        for t in self.purchase:
            t.moveToPixel( (xoff+ x*(BLOCK_WIDTH+margin),yoff+y*(BLOCK_WIDTH+margin)))
            x+=1
            if(x==rowmax):
                x=0
                y=y+1

    def drawBackground(self):
        self.surf.blit(self.bg,(0,0))
        endx = self.width*BLOCK_WIDTH+X_OFFSET
        endy = self.height*BLOCK_WIDTH+Y_OFFSET

        pygame.draw.line(self.surf,self.gColor,(X_OFFSET,endy),(endx,endy),5)
        pygame.draw.line(self.surf,self.gColor,(X_OFFSET,Y_OFFSET),(endx,Y_OFFSET),5)
        pygame.draw.line(self.surf,self.gColor,(X_OFFSET,Y_OFFSET),(X_OFFSET,endy),5)
        pygame.draw.line(self.surf,self.gColor,(endx,Y_OFFSET),(endx,endy),5)

        

        if(self.showGrid):
            for x in range(self.width+1):
                pygame.draw.line(self.surf,self.gColor,(x*BLOCK_WIDTH+X_OFFSET,Y_OFFSET),(x*BLOCK_WIDTH+X_OFFSET,endy))

            for y in range(self.height+1):
                pygame.draw.line(self.surf,self.gColor,(X_OFFSET,y*BLOCK_WIDTH+Y_OFFSET),(endx,y*BLOCK_WIDTH+Y_OFFSET))

        
        if(self.showPolicy):
            if(self.focus != None):
                self.focus.policy.drawDirs(self.surf)
                if(self.legalGridPlacement(gridToPixel(self.focus.getPos()) )\
                   and self.state != Game.POSITIONING):
                      self.focus.drawRange(self.surf)
            elif( self.lastPolicy!=None ):
                self.lastPolicy.drawDirs(self.surf)              
                #self.towerPolicy.drawDirs(self.surf)                                        


    hudFG=pygame.Color(200,200,200)
    hudBG=pygame.Color(155,155,155)

    def drawEntityData(self,x,y):
        if(self.focus !=None):
            self.focus.drawDescription(self.surf,x,y)
            return

    def legalGridPlacement(self,pos):
        """pos is in pixel coords"""
        x,y = pixelToGrid(pos)
        if(0<=x<self.width and 0<=y<self.height):
            #check that we are not colliding with any of the existing towers
            for t in self.towers.sprites():
                if(t.containsPoint( pos) ):
                    return False
            return True
        else:
            return False

    hud = pygame.image.load("tray.png")
    hud2 = pygame.image.load("tray2.png")
    bullet_on = pygame.image.load("tray2-bullet.png")
    health_on = pygame.image.load("tray2-health.png")
    grid_on = pygame.image.load("tray2-grid.png")
    policy_on = pygame.image.load("tray2-policy.png")

    def drawHud(self):

        font = pygame.font.Font("ARNPRIOR.TTF",14)

        #handle the preferences
        pos = (X_OFFSET,390)
        self.surf.blit(Game.hud2,pos)
        if(self.showPolicy) : self.surf.blit(Game.policy_on,pos)
        if(self.showBullets): self.surf.blit(Game.bullet_on,pos)
        if(self.showHealth) : self.surf.blit(Game.health_on,pos)
        if(self.showGrid)   : self.surf.blit(Game.grid_on,pos)
        
        waveSurface = font.render("%d"%(self.waveNumber),False,(0,0,0))
        self.surf.blit(waveSurface,(44,422))

 
        scoreSurface = font.render("%d"%(self.score),False,(0,0,0))
        self.surf.blit(scoreSurface,(116,422))



        #base
        rec = (420,20 ,200,440)
        #pygame.draw.rect( self.surf,Game.hudBG,rec)

        #image background
        self.surf.blit(Game.hud,rec)



        #title
        rec = (430,30,180,40)
        ##pygame.draw.rect( self.surf,Game.hudFG,rec)

        #towers
        rec = (430,80,180,100)



        ##pygame.draw.rect( self.surf,Game.hudFG,rec)
        for t in self.purchase:
            if (t.getCost() < self.money):
                t.setImageState(Entity.IMG_ATTACKING)
            else:
                t.setImageState(Entity.IMG_DISABLED)
            
            self.surf.blit(t.image, t.rect)

        #display
        rec = (430,190,180,200)
        ##pygame.draw.rect( self.surf,Game.hudFG,rec)

        self.drawEntityData(440,200)
       
        


        #cash
        font = pygame.font.Font("ARNPRIOR.TTF",18)
        messageSurface = font.render("$%d"%(self.money),False,(0,0,0))
        self.surf.blit(messageSurface,(440,410))

        if(self.state==Game.POSITIONING):
            pos = pygame.mouse.get_pos()
            self.newTower.moveTLTo(pixelToIntGrid(pos))
            if(self.legalGridPlacement(pos)):
                self.newTower.setImageState(Entity.IMG_ATTACKING)
                self.newTower.drawRange(self.surf)
            else:
                self.newTower.setImageState(Entity.IMG_DISABLED)
            #print "tower at", self.newTower.getPos()
            self.surf.blit(self.newTower.image, self.newTower.rect)
            
    def handlePreferences(self,pos,buttons):
        if(buttons[0]):
            x = pos[0] - X_OFFSET
            y = pos[1] - 390
            if( 34<y<56):
                if(x>380):
                    return
                elif(x>326):
                    self.showBullets=not self.showBullets
                elif(x>268):
                    self.showHealth=not self.showHealth
                elif(x>217):
                    self.showPolicy=not self.showPolicy
                elif(x>180):
                    self.showGrid=not self.showGrid
                       
    def handleMouse(self):
        buttons = pygame.mouse.get_pressed()
        pos = pygame.mouse.get_pos()
        self.handlePreferences(pos,buttons)
        
        if(buttons[2]):#right click?
            if(self.state==Game.POSITIONING):
                self.money+=self.newTower.getCost()

            self.newTower=None
            self.state=Game.HOVER
            self.focus=None
            return
        if (not buttons[0] and (self.state==Game.SELECTED or self.state==Game.POSITIONING) ):
            #nothing to do
            return
        if (self.state==Game.POSITIONING): #button is pressed!
            if(self.legalGridPlacement(pos)):
                self.newTower.moveTLTo(pixelToIntGrid(pos))
                self.plantTower(self.newTower)
                self.focus = self.newTower
                self.newTower = None
                self.state=Game.HOVER
            return
        #we are hovering or buying
        self.focus = None

        for i in self.purchase:
            if(i.containsPoint(pos)):
                 self.focus = i
                 if(buttons[0]):#you bought it! revisit when ee have money
                    self.newTower = i.clone()
                    self.newTower.addPolicy(self.towerPolicy)
                    self.focus=self.newTower
                    #self.tF.getTower("light",pixelToGrid(pos))
                    self.money -= self.newTower.getCost()
                    self.state=self.POSITIONING
                    return

        #check if we are selecting something now
        entities = self.towers.sprites()+self.robots.sprites()
        for i in entities:
            if(i.containsPoint(pos)):
                 self.focus = i
                 if(buttons[0]):
                    self.state=self.SELECTED
                    return                     

    def jitterPos(self,pos):
       dx = random.randint(-1*BLOCK_WIDTH/2,0)/float(BLOCK_WIDTH)
       dy = random.randint(-1*BLOCK_WIDTH/2,0)/float(BLOCK_WIDTH)
       return (pos[0]+dx,pos[1]+dy)

         
    def doSplashScreen(self,filename):
        i = pygame.image.load(filename)
        self.surf.blit(i,(50,100))
        pygame.display.update()
        time.sleep(1)
        while(True):

            for event in pygame.event.get():
                if event.type==pygame.KEYDOWN or event.type == pygame.MOUSEBUTTONUP:
                    return
            time.sleep(.1)



    def run(self):
        self.doSplashScreen("splash.png")
        wave = self.waves[0]
        
        hpmod = 1
        while(self.base.currentHP>0):

            if(self.waveNumber >20 and len(self.robots.sprites()) ==0):
                    self.doSplashScreen("splash2.png")
                    return


            #waves
            if(wave != None and wave.isDone()):
                hpmod=hpmod*1.1
                #print hpmod
                self.waves.remove(wave)
                wave = None
                if(len(self.waves)>0):
                    wave = self.waves[0]
                    #print wave.rType
                    self.robotPolicy=None

            if(wave !=None):       
                rt = wave.doAction()
            else:
                rt = None
            
            if(rt != None):
                robot = self.rF.getRobot(rt,self.jitterPos(self.launch.getPos()),hpmod)
                newBar = Healthbar(robot)
                self.healthbars.add(newBar)
                if(self.robotPolicy==None):
                    self.robotPolicy = Policy(self.width,self.height,1,robot,self.towers.sprites(),self.base.getPos())
                    self.policies.add(self.robotPolicy)
                    self.lastPolicy = self.robotPolicy
                    self.waveNumber+=1

                robot.addPolicy(self.robotPolicy)
                robot.setGame(self)
                self.robots.add(robot)
                

            #updates
            self.robots.update(self.robots.sprites())
            self.towers.update(self.projectiles,self.robots.sprites(),self.surf,self.healthbars)
            self.projectiles.update()
            self.healthbars.update()

            self.handleMouse()
            pygame.event.pump()

            #draw
            self.drawBackground()
            self.towers.draw(self.surf)
            self.robots.draw(self.surf)
            if(self.showHealth):  self.healthbars.draw(self.surf)
            if(self.showBullets): self.projectiles.draw(self.surf)
            self.drawHud()
            pygame.display.update()
            time.sleep(.05)
        self.doSplashScreen("splashlose.png")



def main():
    pygame.init()
    g = Game()
    g.run()
    pygame.quit()

main()
          
